package com.framsticks.communication;

import com.framsticks.communication.queries.*;
import com.framsticks.params.SourceInterface;
import com.framsticks.util.FramsticksException;
import com.framsticks.util.lang.Holder;
import com.framsticks.util.lang.Pair;
import com.framsticks.util.lang.Strings;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import java.net.Socket;
import com.framsticks.util.dispatching.RunAt;

/**
 * @author Piotr Sniegowski
 */
public class ServerConnection extends Connection {

	private final static Logger log = Logger.getLogger(ServerConnection.class);

	RequestHandler requestHandler;

	public ServerConnection(Socket socket, RequestHandler requestHandler) {
		super(socket.getInetAddress().getHostAddress(), "server connection");
		this.socket = socket;
		this.requestHandler = requestHandler;
		// socket.setSoTimeout(500);
		setupStreams();
	}


	@Override
	protected void receiverThreadRoutine() {
		while (isRunning() && isConnected()) {
			try {
				processNextRequest();
			} catch (Exception e) {
				log.log((isRunning() ? Level.ERROR : Level.DEBUG), "caught exception: ", e);
				break;
			}
		}
		interrupt();
		finish();
	}

	protected void handleRequest(Request request, ResponseCallback<?> responseCallback) {
		if (request instanceof ApplicationRequest) {
			requestHandler.handle((ApplicationRequest) request, responseCallback);
			return;
		}
		if (request instanceof ProtocolRequest) {
			if (request instanceof VersionRequest) {
				responseCallback.process(new Response(true, null, null));
				return;
			}
			if (request instanceof UseRequest) {
				String feature = ((UseRequest)request).getFeature();
				if (feature.equals("request_id")) {
					requestIdEnabled = true;
					responseCallback.process(new Response(true, null, null));
					return;
				}
				responseCallback.process(new Response(false, "\"unknown feature: " + feature + "\"", null));
				return;
			}

		}
		log.error("unhandled request: " + request);
		responseCallback.process(new Response(false, "unhandled", null));
	}

	protected final void respond(final Response response, final Integer id) {
		senderThread.dispatch(new RunAt<Connection>() {
			@Override
			public void run() {
				String outId = id != null ? " " + id : "";
				if (response.getFiles() != null) {
					for (File f : response.getFiles()) {
						output.print("file" + outId/* + " " + f.getPath()*/ + "\n");
						SourceInterface content = f.getContent();
						String line;
						while ((line = content.readLine()) != null) {
							output.print(line);
							output.print('\n');
						}
						output.print("eof\n");
					}
				}
				output.print(response.getOk() ? "ok" : "error");
				output.print(outId);
				if (Strings.notEmpty(response.getComment())) {
					output.print(" \"");
					output.print(response.getComment());
					output.print('"');
				}
				output.print('\n');
				output.flush();
			}
		});

	}

	protected void processNextRequest() {
		final Holder<Integer> id = new Holder<>();
		final String line = getLine();
		try {
			Pair<CharSequence, CharSequence> command = Request.takeIdentifier(line);
			final Pair<Integer, CharSequence> rest = takeRequestId(command.second);
			id.set(rest.first);

			final Request request = Request.parse(command.first, rest.second);

			if (log.isTraceEnabled()) {
				log.trace("read request: " + request);
			}

			handleRequest(request, new ResponseCallback<ServerConnection>() {
				@Override
				public void process(Response response) {
					respond(response, rest.first);
				}
			});
		} catch (FramsticksException e) {
			e.arg("id", id.get()).arg("line", line);
			log.error("error: ", e);
			respond(new Response(false, "invalid input: " + e.getMsg(), null), id.get());
			return;
		}

	}
}
