package com.framsticks.core;

import org.apache.log4j.Logger;

import com.framsticks.params.AccessInterface;
import com.framsticks.params.CompositeParam;
import com.framsticks.params.FramsClass;
import com.framsticks.params.Param;
import com.framsticks.params.ValueParam;
import com.framsticks.params.annotations.AutoAppendAnnotation;
import com.framsticks.params.annotations.FramsClassAnnotation;
import com.framsticks.util.UnsupportedOperationException;
import com.framsticks.params.types.ProcedureParam;
import com.framsticks.util.FramsticksException;
import com.framsticks.util.StateFunctor;
import com.framsticks.util.dispatching.Future;
// import static com.framsticks.core.InstanceUtils.*;

@FramsClassAnnotation
public class ObjectInstance extends AbstractInstance {
	private static final Logger log = Logger.getLogger(ObjectInstance.class);

	@AutoAppendAnnotation
	public void setRootObject(Object object) {
		registry.registerAndBuild(object.getClass());
		AccessInterface access = registry.createAccess(object.getClass());
		setRoot(new Node(Param.build().forAccess(access).id(getName()).finish(CompositeParam.class), object));
	}

	public Object getRootObject() {
		return getRoot().getObject();
	}

	public <T> T getRootObject(Class<T> type) {
		Object result = getRootObject();
		if (result == null) {
			throw new FramsticksException().msg("object instance is empty").arg("instance", this);
		}
		if (!type.isInstance(result)) {
			throw new FramsticksException().msg("object instance holds object of different kind").arg("object", result).arg("requested", type).arg("instance", this);
		}
		return type.cast(result);
	}

	@Override
	public void fetchValues(final Path path, final StateFunctor stateFunctor) {
		assert isActive();
		log.debug("requesting: " + path);
		fireFetch(path);
		stateFunctor.call();
	}

	@Override
	public void fetchValue(Path path, ValueParam param, StateFunctor stateFunctor) {
		assert isActive();
		fireFetch(path);
		stateFunctor.call();
	}

	@Override
	public void call(Path path, ProcedureParam param, Object[] arguments, Future<Object> future) {
		assert isActive();
		try {
			future.pass(InstanceUtils.bindAccess(path).call(param, arguments));
		} catch (FramsticksException e) {
			future.handle(e);
		}
	}

	@Override
	public void fetchInfo(Path path, Future<FramsClass> future) {
		assert isActive();
		Path p = path.tryResolveIfNeeded();
		Class<?> javaClass = p.getTopObject().getClass();
		FramsClass framsClass = registry.registerReflectedIfNeeded(javaClass);
		if (framsClass != null) {
			future.pass(framsClass);
		} else {
			future.handle(new FramsticksException().msg("failed to find info for class").arg("java class", javaClass));
		}
	}

	@Override
	public void resolve(Path path, Future<Path> future) {
		assert isActive();
		assert path.isOwner(this);
		if (path.getTop().getObject() != null) {
			future.pass(path);
			return;
		}
		AccessInterface access = InstanceUtils.bindAccess(this, path.getUnder());
		Object object = access.get(path.getTop().getParam(), Object.class);
		if (object == null) {
			future.pass(path);
			return;
		}
		future.pass(path.appendResolution(object));
	}

	@Override
	public void storeValue(Path path, ValueParam param, Object value, final StateFunctor stateFunctor) {
		assert isActive();
		stateFunctor.handle(new UnsupportedOperationException());
	}

	@Override
	public Path create(Path path) {
		assert isActive();
		assert !path.isResolved();
		throw new UnsupportedOperationException();
	}

}
