package com.framsticks.remote;

import com.framsticks.communication.*;
import com.framsticks.communication.queries.CallRequest;
import com.framsticks.communication.queries.GetRequest;
import com.framsticks.communication.queries.InfoRequest;
import com.framsticks.communication.queries.SetRequest;
import com.framsticks.core.AbstractInstance;
import com.framsticks.core.InstanceUtils;
import com.framsticks.core.ListChange;
import com.framsticks.core.Node;
import com.framsticks.core.Path;
import com.framsticks.params.*;
import com.framsticks.params.annotations.FramsClassAnnotation;
import com.framsticks.params.annotations.ParamAnnotation;
import com.framsticks.params.types.EventParam;
import com.framsticks.params.types.ProcedureParam;
import com.framsticks.parsers.MultiParamLoader;
import com.framsticks.core.Instance;
import com.framsticks.util.*;
import com.framsticks.util.dispatching.Dispatching;
import com.framsticks.util.dispatching.ExceptionResultHandler;
import com.framsticks.util.dispatching.Future;
import com.framsticks.util.dispatching.Joinable;
import com.framsticks.util.dispatching.JoinableParent;
import com.framsticks.util.dispatching.JoinableState;
import com.framsticks.util.lang.Casting;
import com.framsticks.util.lang.Pair;
import com.framsticks.util.dispatching.RunAt;
import static com.framsticks.core.InstanceUtils.*;

import java.util.*;

import javax.annotation.Nonnull;

import org.apache.log4j.Logger;

/**
 * @author Piotr Sniegowski
 */
@FramsClassAnnotation
public class RemoteInstance extends AbstractInstance implements JoinableParent {

	private final static Logger log = Logger.getLogger(RemoteInstance.class.getName());

	protected ClientConnection connection;

	protected final Set<Pair<Path, Subscription<?>>> subscriptions = new HashSet<>();

	public Pair<Path, Subscription<?>> getSubscription(Path path) {
		for (Pair<Path, Subscription<?>> s : subscriptions) {
			if (s.first.matches(path)) {
				return s;
			}
		}
		return null;
	}

	public RemoteInstance() {
	}

	@ParamAnnotation
	public void setAddress(String address) {
		setConnection(new ClientConnection(address));
	}

	@ParamAnnotation
	public String getAddress() {
		return connection == null ? "<disconnected>" : connection.getAddress();
	}

	protected void onProtocolVersionNegotiated() {
	}


	public void setConnection(final ClientConnection connection) {
		this.connection = connection;
		final ExceptionResultHandler failure = new ExceptionResultHandler() {
			@Override
			public void handle(FramsticksException exception) {
				log.fatal("failed to establish connection: ", exception);
				// log.fatal("unsupported protocol version!\n minimal version is: " + "\nmanager protocol is: " + connection.getProtocolVersion());
				Dispatching.drop(connection, RemoteInstance.this);
				fireRun(exception);
			}
		};

		this.connection.setConnectedFunctor(new AbstractStateFunctor(failure) {
			@Override
			public void call() {
				connection.negotiateProtocolVersion(new AbstractStateFunctor(failure) {
					@Override
					public void call() {
						onProtocolVersionNegotiated();
					}
				});
			}
		});
	}

	@Override
	public String toString() {
		assert Dispatching.isThreadSafe();
		return "remote instance " + getName() + "(" + getAddress() + ")";
	}

	public final ClientConnection getConnection() {
		return connection;
	}

	@Override
	public void fetchValue(final Path path, final ValueParam param, final StateFunctor stateFunctor) {
		assert isActive();
		assert param != null;
		assert path.isResolved();
		connection.send(new GetRequest().field(param.getId()).path(path.getTextual()), this, new ResponseCallback<Instance>() {
			@Override
			public void process(Response response) {
				assert isActive();
				if (!response.getOk()) {
					stateFunctor.handle(new FramsticksException().msg("failed to fetch value").arg("comment", response.getComment()));
					return;
				}
				try {
					InstanceUtils.processFetchedValues(path, response.getFiles());
					stateFunctor.call();
				} catch (FramsticksException ex) {
					stateFunctor.handle(ex);
				}
			}
		});
	}

	protected final Map<String, Set<Future<FramsClass>>> infoRequests = new HashMap<String, Set<Future<FramsClass>>>();

	protected void finishInfoRequest(String id, FramsClass result, FramsticksException e) {
		assert isActive();
		Set<Future<FramsClass>> futures = infoRequests.get(id);
		infoRequests.remove(id);
		for (Future<FramsClass> f : futures) {
			Future.passOrHandle(f, result, e);
		}
	}

	@Override
	public void fetchInfo(final Path path, final Future<FramsClass> future) {

		final String name = path.getTop().getParam().getContainedTypeName();

		if (infoRequests.containsKey(name)) {
			infoRequests.get(name).add(future);
			return;
		}

		log.debug("issuing info request for " + name);
		Set<Future<FramsClass>> futures = new HashSet<Future<FramsClass>>();
		futures.add(future);
		infoRequests.put(name, futures);

		//TODO: if the info is in the cache, then don't communicate
		connection.send(new InfoRequest().path(path.getTextual()), this, new ResponseCallback<Instance>() {
			@Override
			public void process(Response response) {
				assert isActive();
				try {
					if (!response.getOk()) {
						throw new FramsticksException().msg("invalid response").arg("comment", response.getComment());
					}
					if (response.getFiles().size() != 1) {
						throw new FramsticksException().msg("invalid number of files in response").arg("files", response.getFiles().size());
					}
					if (!path.isTheSame(response.getFiles().get(0).getPath())) {
						throw new FramsticksException().msg("path mismatch").arg("returned path", response.getFiles().get(0).getPath());
					}
					FramsClass framsClass = InstanceUtils.processFetchedInfo(RemoteInstance.this, response.getFiles().get(0));

					CompositeParam thisParam = path.getTop().getParam();
					if (!thisParam.isMatchingContainedName(framsClass.getId())) {
						throw new FramsticksException().msg("framsclass id mismatch").arg("request", thisParam.getContainedTypeName()).arg("fetched", framsClass.getId());
					}

					finishInfoRequest(name, framsClass, null);
				} catch (FramsticksException e) {
					finishInfoRequest(name, null, e.arg("path", path));
				}
			}
		});
	}

	@Override
	public void fetchValues(final Path path, final StateFunctor stateFunctor) {
		assert isActive();
		assert path.getTop().getObject() != null;

		log.trace("fetching values for " + path);
		connection.send(new GetRequest().path(path.getTextual()), this, new ResponseCallback<Instance>() {
			@Override
			public void process(Response response) {
				assert isActive();
				try {
					if (!response.getOk()) {
						throw new FramsticksException().msg("failed to fetch values").arg("comment", response.getComment());
					}
					InstanceUtils.processFetchedValues(path, response.getFiles());
					stateFunctor.call();
				} catch (FramsticksException e) {
					stateFunctor.handle(e);
				}
			}
		});
	}

	@Override
	public void resolve(final Path path, final Future<Path> future) {
		InstanceUtils.resolve(path, future);
	}

	@Override
	protected void tryRegisterOnChangeEvents(final Path path) {
		assert isActive();
		AccessInterface access = InstanceUtils.bindAccess(path);
		if (!(access instanceof ListAccess)) {
			return;
		}

		assert path.size() >= 2;
		FramsClass underFramsClass = getInfoFromCache(path.getUnder().getParam().getContainedTypeName());

		EventParam changedEvent;
		try {
			changedEvent = underFramsClass.getParamEntry(path.getTop().getParam().getId() + "_changed", EventParam.class);
		} catch (FramsticksException e) {
			return;
		}

		log.debug("registering for " + changedEvent);
		if (getSubscription(path) != null) {
			return;
		}

		final Pair<Path, Subscription<?>> temporary = new Pair<>(path, null);
		subscriptions.add(temporary);

		connection.subscribe(path.getTextual() + "_changed", this, new SubscriptionCallback<Instance>() {
			@Override
			public EventCallback subscribed(final Subscription<? super Instance> subscription) {
				if (subscription == null) {
					log.error("failed to subscribe for change event for " + path);
					return null;
				}
				log.debug("subscribed for change event for " + path);
				// subscription.setDispatcher(RemoteInstance.this);
				RemoteInstance.this.dispatch(new RunAt<Instance>() {
					@Override
					public void run() {
						subscriptions.remove(temporary);
						subscriptions.add(new Pair<Path, Subscription<?>>(path, subscription));
					}
				});
				return new EventCallback() {
					@Override
					public void call(List<File> files) {
						assert isActive();
						assert files.size() == 1;
						MultiParamLoader loader = new MultiParamLoader();
						loader.setNewSource(files.get(0).getContent());
						loader.addBreakCondition(MultiParamLoader.Status.AfterObject);
						ReflectionAccess access = new ReflectionAccess(ListChange.class, FramsClass.build().forClass(ListChange.class));
						loader.addAccessInterface(access);
						MultiParamLoader.Status status;
						try {
							while ((status = loader.go()) != MultiParamLoader.Status.Finished) {
								if (status == MultiParamLoader.Status.AfterObject) {
									AccessInterface accessInterface = loader.getLastAccessInterface();
									reactToChange(path, (ListChange) accessInterface.getSelected());
								}
							}
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				};
			}
		});
	}

	protected Future<Path> futureListChanger(final ListChange listChange, final String path) {
		return new Future<Path>(Logging.logger(log, "failed to " + listChange, path)) {
			@Override
			protected void result(Path result) {
				log.debug(listChange + ": " + result);
				fireListChange(result, listChange);
			}
		};
	}

	protected void reactToChange(final Path path, final ListChange listChange) {
		assert isActive();
		log.debug("reacting to change " + listChange + " in " + path);
		AccessInterface access = InstanceUtils.bindAccess(path);
		assert access != null;

		if ((listChange.getAction() == ListChange.Action.Modify) && (listChange.getPosition() == -1)) {
			final String p = path.getTextual();
			InstanceUtils.resolveAndFetch(this, p, futureListChanger(listChange, p));
			return;
		}

		CompositeParam childParam = Casting.tryCast(CompositeParam.class, access.getParam(listChange.getBestIdentifier()));
		assert childParam != null;
		switch (listChange.getAction()) {
			case Add: {
				final String p = path.getTextual() + "/" + childParam.getId();
				InstanceUtils.resolveAndFetch(this, p, futureListChanger(listChange, p));
				break;
			}
			case Remove: {
				access.set(childParam, null);
				fireListChange(path, listChange);
				break;
			}
			case Modify: {
				final String p = path.getTextual() + "/" + childParam.getId();
				InstanceUtils.resolveAndFetch(this, p, futureListChanger(listChange, p));
				break;
			}
		}
	}

	@Override
	public void storeValue(final Path path, final ValueParam param, final Object value, StateFunctor stateFunctor) {
		assert isActive();

		log.trace("storing value " + param + " for " + path);
		connection.send(new SetRequest().value(value.toString()).field(param.getId()).path(path.getTextual()), this, new StateCallback<Instance>(stateFunctor) {
			@Override
			protected void callImpl() {
				InstanceUtils.bindAccess(path).set((ValueParam) param, value);
			}
		});
	}

	@Override
	protected void joinableStart() {
		Dispatching.use(connection, this);
		super.joinableStart();
	}

	@Override
	protected void joinableInterrupt() {
		Dispatching.drop(connection, this);
		super.joinableInterrupt();
	}

	@Override
	protected void joinableFinish() {
		super.joinableFinish();

	}

	@Override
	public void joinableJoin() throws InterruptedException {
		Dispatching.join(connection);
		super.joinableJoin();
	}

	@Override
	public void childChangedState(Joinable joinable, JoinableState state) {
		proceedToState(state);
	}

	@Override
	public void call(@Nonnull final Path path, @Nonnull final ProcedureParam procedure, @Nonnull Object[] arguments, final Future<Object> future) {
		assert isActive();
		assert path.isResolved();

		//TODO validate arguments type using params
		connection.send(new CallRequest().procedure(procedure.getId()).arguments(Arrays.asList(arguments)).path(path.getTextual()), this, new ResponseCallback<Instance>() {
			@Override
			public void process(Response response) {
				assert isActive();
				try {
					if (!response.getOk()) {
						throw new FramsticksException().msg("failed to call procedure").arg("procedure", procedure).arg("comment", response.getComment());
					}
					// InstanceUtils.processFetchedValues(path, response.getFiles());
					future.pass(null);
				} catch (FramsticksException ex) {
					future.handle(ex);
				}
			}
		});

	}

	@Override
	public Path create(Path path) {
		assert isActive();
		assert !path.isResolved();
		Path resolved = path.tryFindResolution();
		if (!resolved.isResolved()) {
			log.debug("creating: " + path);
			AccessInterface access = registry.prepareAccess(path.getTop().getParam());
			assert access != null;
			Object child = access.createAccessee();
			assert child != null;
			if (path.size() == 1) {
				setRoot(new Node(getRoot().getParam(), child));
			} else {
				bindAccess(this, path.getUnder()).set(path.getTop().getParam(), child);
			}
			resolved = path.appendResolution(child);
		}
		tryRegisterOnChangeEvents(resolved);
		return resolved;
	}

}
